with TW_Sorting;

with Ada.Text_IO;
with Ada.Numerics.Float_Random;
with Ada.Calendar;
with Ada.Exceptions;
with Ada.Unchecked_Deallocation;

procedure Sort_Test is

   type Array_Type is array (Integer range <>) of Integer;
   type Array_Ptr  is access all Array_Type;

   --  This is just a simple, normal quicksort with median-of-three. It is
   --  included for comparison purposes only.

   procedure Quicksort
     (To_Sort : in out Array_Type)
   is
      Cut_Off : constant := 10;

      Pivot, Temp : Integer;

      procedure Swap (Left, Right : in Integer);
      pragma Inline (Swap); --  inline this for increased performance.

      procedure Swap (Left, Right : in Integer)
      is
      begin
         Temp            := To_Sort (Left);
         To_Sort (Left)  := To_Sort (Right);
         To_Sort (Right) := Temp;
      end Swap;

      procedure Insertion_Sort
        (L, R : in Integer)
      is
         --  Precondition: L < R.
         J : Integer;
      begin --  Insertion_Sort
         for I in Integer range L + 1 .. R loop
            Temp := To_Sort (I);
            J := I;
            while J > L and then Temp < To_Sort (J - 1) loop
               To_Sort (J) := To_Sort (J - 1);
               J := J - 1;
            end loop;
            To_Sort (J) := Temp;
         end loop;
      end Insertion_Sort;

      procedure Sort
        (L, R : in Integer)
      is
         Left   : Integer := L;
         Right  : Integer := R;
         I      : Integer := Left;
         J      : Integer := Right;
         Middle : Integer;

      begin --  Intro_Sort
         while Right - Left > Cut_Off loop
            Middle := Left + (Right - Left) / 2;
            --  Median-of-three:
            if To_Sort (Middle) < To_Sort (Left) then
               Swap (Middle, Left);
            end if;
            if To_Sort (Right) < To_Sort (Left) then
               Swap (Right, Left);
            end if;
            if To_Sort (Right) < To_Sort (Middle) then
               Swap (Right, Middle);
            end if;
            Pivot := To_Sort (Middle);
            --  Here, I = Left and J = Right
            while I < J loop
               while To_Sort (I) < Pivot loop I := I + 1; end loop;
               while Pivot < To_Sort (J) loop J := J - 1; end loop;
               if I <= J then
                  if I < J then Swap (I, J); end if;
                  if I < Integer'Last  then I := I + 1; end if;
                  if J > Integer'First then J := J - 1; end if;
               end if;
            end loop;
            if J - Left > Right - I then
               --  Left part is longer.
               if I < Right then Sort (I, Right); end if;
               I := Left; Right := J;
            else
               --  Right part is longer.
               if J > Left then Sort (Left, J); end if;
               J := Right; Left := I;
            end if;
         end loop;
         if Left < Right then Insertion_Sort (Left, Right); end if;
      end Sort;

   begin --  Quicksort
      if To_Sort'Length > 1 then
         Sort (To_Sort'First, To_Sort'Last);
      end if;
   end Quicksort;

   type Bool_Array is array (Boolean range <>) of Integer;
   type Bool_Ptr   is access all Bool_Array;

   AP   : Array_Ptr;
   BP   : Bool_Ptr;

   Temp : Array_Type (-2 .. -1);

   function A_Is_Smaller (L, R : Integer) return Boolean
   is
      Left, Right : Integer;
   begin
      if L < 0 then
         Left  := Temp (L);
      else
         Left  := AP (L + AP'First);
      end if;
      if R < 0 then
         Right := Temp (R);
      else
         Right := AP (R + AP'First);
      end if;
      return Left < Right;
   end A_Is_Smaller;

   procedure A_Copy (To, From : in Integer)
   is
   begin
      if To < 0 then
         Temp (To) := AP (From + AP'First);
      else
         if From < 0 then
            AP (To + AP'First) := Temp (From);
         else
            AP (To + AP'First) := AP (From + AP'First);
         end if;
      end if;
   end A_Copy;

   function B_Is_Smaller (L, R : Integer) return Boolean
   is
      Left, Right : Integer;
   begin
      if L < 0 then Left  := Temp (L);
               else Left  := BP (Boolean'Val (L));
      end if;
      if R < 0 then Right := Temp (R);
               else Right := BP (Boolean'Val (R));
      end if;
      return Left < Right;
   end B_Is_Smaller;

   procedure B_Copy (To, From : in Integer)
   is
   begin
      if To < 0 then
         Temp (To) := BP (Boolean'Val (From));
      else
         if From < 0 then
            BP (Boolean'Val (To)) := Temp (From);
         else
            BP (Boolean'Val (To)) := BP (Boolean'Val (From));
         end if;
      end if;
   end B_Copy;

   procedure QS  is new TW_Sorting.Sort_G (Integer, Integer, Array_Type);
   procedure BQS is new TW_Sorting.Sort_G (Boolean, Integer, Bool_Array);

   procedure PS  is new TW_Sorting.Sort_Indexed_G (A_Is_Smaller, A_Copy);
   procedure BPS is new TW_Sorting.Sort_Indexed_G (B_Is_Smaller, B_Copy);

   --  Wrap the indexed versions to get a common interface for testing.

   procedure Free is
      new Ada.Unchecked_Deallocation (Array_Type, Array_Ptr);

   procedure PSW (A : in out Array_Type)
   is
   begin
      AP := new Array_Type'(A);
      if A'Length = 0 then
         PS (1, 0);
      else
         PS (0, A'Length-1);
      end if;
      A := AP.all;
      Free (AP);
   end PSW;

   procedure BPSW (A : in out Bool_Array)
   is
      B : aliased Bool_Array := A;
   begin
      BP := B'Unchecked_Access;
      BPS (0, B'Length-1);
      A := B;
   end BPSW;

   type Sorter  is access procedure (A : in out Array_Type);
   type BSorter is access procedure (A : in out Bool_Array);

   type ISorter  is access procedure (L, R : in Natural);
   type BISorter is access procedure (L, R : in Natural);

   type Test is
      record
         Name      : String (1 .. 9);
         Cumulated : Duration;
         Sort      : Sorter;
         ISort     : ISorter;
         BSort     : BSorter;
         BISort    : BISorter;
         Failures  : Natural;
      end record;

   type Test_Table is array (Positive range <>) of Test;

   Tests : Test_Table :=
     (("Sort_G   ", 0.0, QS'Access, null, BQS'Access, null, 0),
      ("Indexed_G", 0.0, null, PS'Access, null, BPS'Access, 0),
      ("Wrapped  ", 0.0, PSW'Access, null, BPSW'Access, null, 0),
      ("QuickSort", 0.0, Quicksort'Access, null, null, null, 0));

   package Rnd renames Ada.Numerics.Float_Random;

   G : Rnd.Generator;

   procedure Randomize
     (A : in out Array_Type)
   is
      --  Fill A with a random shuffle of its index range. A shuffle has the
      --  advantage that we can even detect nonsense behavior. After sorting,
      --  A (I) = I for all I. If we just filled the array with random values
      --  and then tested that A(I) <= A(I+1) for all I in L .. R-1 after
      --  sorting, a stupid implementation that simply set all elements to
      --  zero would be considered correct! Hence the slightly more elaborate
      --  shuffle, which detects such nonsense as a wrong sort.
      N   : Natural;
      R   : Integer;
      Tmp : Integer;
   begin
      for I in A'Range loop
         A (I) := I;
      end loop;
      --  Now shuffle them.
      N := A'Length;
      while N > 1 loop
         R   :=
           Integer (A'First + Integer (Float (N) * Rnd.Random (G)) mod N);
         Tmp := A (A'First +  N - 1);
         A (A'First +  N - 1) := A (R);
         A (R) := Tmp;
         N := N - 1;
      end loop;
   end Randomize;

   procedure Print_Array
     (S : in String;
      A : in Array_Type;
      I : in Integer)
   is
      Low, High : Integer;
   begin
      Low  := Integer'Max (I - 100, A'First);
      if A'Last > 200 and then A'Last - 200 >= Low then
         High := Low + 200;
      else
         High := A'Last;
      end if;
      Ada.Text_IO.Put_Line (S);
      Ada.Text_IO.Put_Line ("Bounds: First =" & Integer'Image (A'First) &
                            " Last =" & Integer'Image (A'Last));
      Ada.Text_IO.Put_Line ("Shown: Low =" & Integer'Image (Low) &
                            " High =" & Integer'Image (High));
      declare
         N : Natural := 0;
      begin
         for J in Low .. High loop
            Ada.Text_IO.Put (Integer'Image (A (J)));
            N := N + 1;
            if N = 10 then
               Ada.Text_IO.New_Line;
               N := 0;
            end if;
         end loop;
         if N /= 0 then Ada.Text_IO.New_Line; end if;
      end;
   end Print_Array;

   procedure Test_Failed
     (S    : in String;
      A, B : in Array_Type;
      I    : in Integer := 0)
   is
   begin
      Ada.Text_IO.Put_Line (">>> TEST FAILED! (" & S & ')');
      Print_Array ("Original:", A, I);
      Ada.Text_IO.Put_Line ("---");
      Print_Array ("Sorted:", B, I);
   end Test_Failed;

   procedure Reset_Timers
   is
   begin
      for I in Tests'Range loop
         Tests (I).Cumulated := 0.0;
      end loop;
   end Reset_Timers;

   procedure Output
   is
   begin
      for I in Tests'Range loop
         Ada.Text_IO.Put_Line ("   " & Tests (I).Name & ':' &
                               Duration'Image (Tests (I).Cumulated));
      end loop;
   end Output;

   procedure Sort_Test
     (T         : in out Test;
      A         : in     Array_Type;
      Verify_It : in     Boolean := True;
      Time_It   : in     Boolean := False)
   is
      use Ada.Calendar;
      B       : aliased Array_Type := A;
      Elapsed : Duration;
      Now     : Time;
   begin
      begin
         if T.Sort /= null then
            Now := Clock;
            T.Sort (B);
            Elapsed := Clock - Now;
         elsif T.ISort /= null then
            AP := B'Unchecked_Access;
            if B'Length = 0 then
               T.ISort (1, 0);
            else
               Now := Clock;
               T.ISort (0, B'Length-1);
               Elapsed := Clock - Now;
            end if;
         else
            return;
         end if;
      exception
         when E : others =>
            Ada.Text_IO.Put_Line ("EXC: " &
                                  Ada.Exceptions.Exception_Information (E));
            Test_Failed ("Raised " & Ada.Exceptions.Exception_Name (E), A, B);
            raise;
      end;
      if Time_It then T.Cumulated := T.Cumulated + Elapsed; end if;
      if Verify_It then
         --  Since the elements of the array are taken from its index range,
         --  B(I) = I for all I.
         for I in B'Range loop
            if B (I) /= I then
               Test_Failed ("Not sorted", A, B, I);
               raise Program_Error;
            end if;
         end loop;
      end if;
   end Sort_Test;

   Num : Natural := 0;

   procedure Sort_It
     (A         : in Array_Type;
      Verify_It : in Boolean := True;
      Time_It   : in Boolean := False)
   is
   begin
      Num := Num + 1;
      for I in Tests'Range loop
         begin
            Sort_Test (Tests (I), A, Verify_It, Time_It);
         exception
            when others =>
               Ada.Text_IO.Put_Line (Tests (I).Name & ": Test" &
                                     Natural'Image (Num) & " FAILED!");
               Tests (I).Failures := Tests (I).Failures + 1;
         end;
      end loop;
   end Sort_It;

   procedure Huge_Test
   is
      A : Array_Ptr := new Array_Type (1 .. 10_000_000);
      procedure Randomize (A : in Array_Ptr)
      is
         N   : Natural;
         R   : Integer;
         Tmp : Integer;
      begin
         for I in A'Range loop
            A (I) := I;
         end loop;
         --  Now shuffle them.
         N := A'Length;
         while N > 1 loop
            R   :=
              Integer (A'First + Integer (Float (N) * Rnd.Random (G)) mod N);
            Tmp := A (A'First +  N - 1);
            A (A'First +  N - 1) := A (R);
            A (R) := Tmp;
            N := N - 1;
         end loop;
      end Randomize;

      procedure Sort_Test
        (T         : in out Test;
         A         : in     Array_Ptr)
      is
         use Ada.Calendar;
         Elapsed : Duration;
         Now     : Time;
      begin
         begin
            if T.Sort /= null then
               Now := Clock;
               T.Sort (A.all);
               Elapsed := Clock - Now;
            elsif T.ISort /= null then
               AP := A;
               Now := Clock;
               T.ISort (0, A'Length-1);
               Elapsed := Clock - Now;
            else
               return;
            end if;
         exception
            when E : others =>
               Ada.Text_IO.Put_Line ("EXC: " &
                                     Ada.Exceptions.Exception_Information (E));
            raise;
         end;
         T.Cumulated := T.Cumulated + Elapsed;
         --  Since the elements of the array are taken from its index range,
         --  A(I) = I for all I.
         for I in A'Range loop
            if A (I) /= I then
               Ada.Text_IO.Put_Line ("not sorted");
               raise Program_Error;
            end if;
         end loop;
      end Sort_Test;

      procedure Sort_It
        (A : in Array_Ptr)
      is
      begin
         Num := Num + 1;
         for I in Tests'Range loop
            begin
               Ada.Text_IO.Put_Line ("Randomizing...");
               Randomize (A);
               Ada.Text_IO.Put_Line ("Sorting...");
               Sort_Test (Tests (I), A);
            exception
               when others =>
                  Ada.Text_IO.Put_Line (Tests (I).Name & ": Test" &
                                        Natural'Image (Num) & " FAILED!");
                  Tests (I).Failures := Tests (I).Failures + 1;
            end;
         end loop;
      end Sort_It;

   begin
      Num := Num + 1;
      Sort_It (A);
   end Huge_Test;

   procedure Test_Boolean
   is
   begin
      Ada.Text_IO.Put_Line ("Boolean tests...");
      declare
         A : Bool_Array (Boolean) := (1000, -1000);
         B : aliased Bool_Array := A;
      begin
         for I in Tests'Range loop
            if Tests (I).BSort /= null or else Tests (I).BISort /= null then
               B := A;
               begin
                  if Tests (I).BSort /= null then
                     Tests (I).BSort (B);
                  else
                     BP := B'Unchecked_Access;
                     Tests (I).BISort (0, 1);
                  end if;
                  if B (False) > B (True) then
                     Ada.Text_IO.Put_Line
                       (Tests (I).Name & " failed boolean case 1: not sorted");
                     Tests (I).Failures := Tests (I).Failures + 1;
                  end if;
               exception
                  when E : others =>
                     Ada.Text_IO.Put_Line
                       ("EXC: " & Ada.Exceptions.Exception_Information (E));
                     Ada.Text_IO.Put_Line
                       (Tests (I).Name & " failed on boolean case 1: " &
                        Ada.Exceptions.Exception_Name (E));
                     Tests (I).Failures := Tests (I).Failures + 1;
               end;
            end if;
         end loop;
      end;
      declare
         A : Bool_Array (False .. False) := (False => -1000);
         B : aliased Bool_Array := A;
      begin
         for I in Tests'Range loop
            if Tests (I).BSort /= null or else Tests (I).BISort /= null then
               B := A;
               begin
                  if Tests (I).BSort /= null then
                     Tests (I).BSort (B);
                  else
                     BP := B'Unchecked_Access;
                     Tests (I).BISort (0, 0);
                  end if;
                  if B (False) /= -1000 then
                     Ada.Text_IO.Put_Line
                       (Tests (I).Name & " failed boolean case 2: not sorted");
                     Tests (I).Failures := Tests (I).Failures + 1;
                  end if;
               exception
                  when E : others =>
                     Ada.Text_IO.Put_Line
                       ("EXC: " & Ada.Exceptions.Exception_Information (E));
                     Ada.Text_IO.Put_Line
                       (Tests (I).Name & " failed on boolean case 2: " &
                        Ada.Exceptions.Exception_Name (E));
                     Tests (I).Failures := Tests (I).Failures + 1;
               end;
            end if;
         end loop;
      end;
      Ada.Text_IO.Put_Line ("Boolean tests completed.");
   end Test_Boolean;

begin
   Rnd.Reset (G);
   Ada.Text_IO.Put_Line ("Simple tests...");
   --  First some simple tests:
   --  Index range at low end.
   declare
      A : Array_Type (1 .. 2) := (1, 2);
   begin
      Sort_It (A);
   end;
   --  Index range at high end.
   declare
      A : Array_Type (Integer'Last - 1 .. Integer'Last) :=
        (Integer'Last, Integer'Last - 1);
   begin
      Sort_It (A);
   end;
   --  Empty array.
   declare
      A : Array_Type (2 .. 1) := (others => 1);
   begin
      Sort_It (A);
   end;
   --  Pre-sorted.
   declare
      A : Array_Type (1 .. 100);
   begin
      for I in A'Range loop
         A (I) := I;
      end loop;
      Sort_It (A);
   end;
   --  Inverted.
   declare
      A : Array_Type (1 .. 100);
   begin
      for I in A'Range loop
         A (I) := 101 - I;
      end loop;
      Sort_It (A);
   end;
   --  All equals.
   declare
      A : Array_Type (1 .. 100) := (others => 50);
   begin
      Sort_It (A, False);
   end;
   --  Random array.
   declare
      A : Array_Type (11 .. 111);
   begin
      Randomize (A);
      Sort_It (A);
   end;
   Ada.Text_IO.Put_Line ("Simple tests done.");
   --  Arrays with negative lower bound:
   declare
      A : Array_Type (Integer'First .. Integer'First + 100);
   begin
      Randomize (A);
      Sort_It (A);
   end;
   --  Arrays with boolean bounds:
   Test_Boolean;
   Ada.Text_IO.New_Line;
   Ada.Text_IO.Put_Line ("Timings:");
   Ada.Text_IO.Put_Line ("--------");
   Ada.Text_IO.New_Line;
   --  OK, now time them.
   declare
      Max : constant Positive := 2000;
   begin
      Ada.Text_IO.Put_Line
        ("Randomized arrays of size 2 .." & Positive'Image (Max));
      Ada.Text_IO.Put_Line ("   (May take some time...)");
      Reset_Timers;
      for I in Positive range 2 .. Max loop
         declare
            A : Array_Type (10 .. 10 + I);
         begin
            Randomize (A);
            Sort_It (A, True, True);
         end;
      end loop;
      Output;
   end;
   --  Now some special cases:
   declare
      Max  : constant Positive := 2 ** 15; -- Must be a power of two!
      A    : Array_Type (1 .. Max);
      N, J : Positive;
   begin
      Ada.Text_IO.Put_Line ("Special cases (size =" &
                            Positive'Image (Max) & ")");
      --  All sorted:
      for I in A'Range loop
         A (I) := I;
      end loop;
      Ada.Text_IO.Put_Line ("Sorted array");
      Reset_Timers;
      Sort_It (A, True, True);
      Output;
      --  Inverted array:
      for I in A'Range loop
         A (I) := Max - I + 1;
      end loop;
      Ada.Text_IO.Put_Line ("Inverted array");
      Reset_Timers;
      Sort_It (A, True, True);
      Output;
      --  All equals:
      for I in A'Range loop
         A (I) := 1000;
      end loop;
      Ada.Text_IO.Put_Line ("All elements = 1000");
      Reset_Timers;
      Sort_It (A, False, True);
      Output;
      --  Middle killer: causes N**2 behavior for quicksorts that take
      --  Pivot := A ((L+R)/2)
      for I in Positive range 1 .. Max / 2 loop
         A (I) := 2 * I;
      end loop;
      N := 2; J := 1;
      while J < Max loop
         A (Max / N *  J + 1) := J;
         J := J + 2;
         if J > N then N := N * 2; end if;
      end loop;
      Ada.Text_IO.Put_Line ("Middle killer");
      Reset_Timers;
      Sort_It (A, True, True);
      Output;
      --  Middle killer K+1: causes N**2 behavior for quicksorts that take
      --  Pivot := A (L + N/2)
      for I in Positive range 1 .. Max / 2 loop
         A (I) := 2 * I - 1;
      end loop;
      N := 2; J := 1;
      while J < Max loop
         if J = Max - 1 then
            A (Max / 2 + 1) := Max;
         else
            A (Max / N * J + 2) := J + 1;
         end if;
         J := J + 2;
         if J > N then N := N * 2; end if;
      end loop;
      Ada.Text_IO.Put_Line ("Middle killer (K+1)");
      Reset_Timers;
      Sort_It (A, True, True);
      Output;
      --  Median-of-three killer: causes N**2 behavior for quicksorts that take
      --  Pivot := A (Median_Of_3 (A (L), A (R), A ((L+R)/2)));
      for I in Positive range 1 .. Max / 2 loop
         A (I) := 2 * I - 1;
      end loop;
      N := 2; J := 1;
      while J < Max loop
         A (Max / N * J + 1) := J + 1;
         J := J + 2;
         if J > N then N := N * 2; end if;
      end loop;
      Ada.Text_IO.Put_Line ("Median-of-three killer");
      Reset_Timers;
      Sort_It (A, True, True);
      Output;
      --  Median-of-three killer (K+1): causes N**2 behavior for quicksorts
      --  that take Pivot := A (Median_Of_3 (A (L), A (R), A (L + N/2)))
      J := 1;
      while J <= Max / 2 loop
         if J mod 2 = 1 then
            A (J) := J;
         else
            A (J) := Max / 2 + J - 1;
         end if;
         J := J + 1;
      end loop;
      N := 2;
      while J <= Max loop
         A (J) := N; N := N + 2; J := J + 1;
      end loop;
      Ada.Text_IO.Put_Line ("Median-of-three killer (K+1)");
      Reset_Timers;
      Sort_It (A, True, True);
      Output;
      --  Now a few large randomized arrays.
      Ada.Text_IO.Put_Line ("50 random arrays (may take some time)...");
      Reset_Timers;
      for I in 1 .. 50 loop
         Randomize (A);
         Sort_It (A, True, True);
      end loop;
      Output;
      --  And now a very large array
      Ada.Text_IO.Put_Line ("Sorting an array with 10 million elements...");
      Reset_Timers;
      Huge_Test;
      Output;
   end;
   Ada.Text_IO.New_Line;
   Ada.Text_IO.Put_Line ("Summary:");
   for I in Tests'Range loop
      if Tests (I).Failures > 0 then
         Ada.Text_IO.Put_Line
           ("- " & Tests (I).Name & " has failed" &
            Natural'Image (Tests (I).Failures) & " tests.");
      else
         Ada.Text_IO.Put_Line
           ("+ " & Tests (I).Name & " has passed all tests!");
      end if;
   end loop;
   Ada.Text_IO.Put_Line ("Quicksort tests done.");
end Sort_Test;
