-------------------------------------------------------------------------------
--
-- Copyright (c) 1999 - 2000 by Thomas Wolf.
--   This piece of software is freeware; it may be redistributed and/or
--   modified under the terms of the "Ada Community License" included in
--   this distribution (see file "ACL.html").
--
-- Disclaimer:
--   This software is provided "as is", without any express or implied
--   warranties, including, without limitation, any implied warranties of
--   merchantibility and fitness for a particular purpose.
--   (In short: use at your own risk :-)
--
-- Author: Thomas Wolf  (TW) twolf@acm.org
--
-- Purpose:
--   Speed and space optimized quicksort. Actually, an introspective
--   quicksort with a worst-case runtime complexity of O(N*log2(N)).
--
-- Literature:
--   Musser, D.R.: "Introspective Sorting and Selection Algorithms",
--      Software -- Practice & Experience (8):983-993; 1997.
--
-- Tasking semantics:
--   N/A. Not abortion-safe.
--
-- Storage semantics:
--   No storage allocation. Stack space used is O(log2(N)).
--
-- Revision History
--
--   21-JAN-1999   TW  Initial version
-------------------------------------------------------------------------------

package TW_Sorting is

   pragma Elaborate_Body;

   ----------------------------------------------------------------------------
   --  A sort with a classic interface:

   generic
      type Index_Type   is (<>);
      type Element_Type is private;
      type Array_Type   is array (Index_Type range <>) of Element_Type;
      with function "<" (Left, Right : in Element_Type) return Boolean is <>;
   procedure Sort_G
     (To_Sort : in out Array_Type);
   --  Sorts the array into ascending order according to the given function
   --  "<". This is an introspective quicksort with average *and* worst-case
   --  performance complexity of O(log2(N)). Stack space usage is bounded by
   --  log2(N).
   --     If To_Sort'Length >= System.Max_Int, Constraint_Error may be raised.
   --  (I didn't test that!) However, System.Max_Int is typically >= 2**31-1,
   --  and sorting arrays of 2 Gigabytes or more is not exactly a common case.

   ----------------------------------------------------------------------------
   --  A very general sort that can be used to sort whatever you like. As
   --  long as you can provide random access in constant time, this will
   --  be a logarithmic sort. (It's an introspective quicksort, too.)

   generic
      with function  Is_Smaller (Left, Right : in Integer) return Boolean;
      --  Shall return True if the element at index 'Left' is smaller than
      --  the element at index 'Right' and Fasle otherwise.
      with procedure Copy       (To, From    : in Integer);
      --  Shall copy the element at index 'From' to position 'To'.
   procedure Sort_Indexed_G
     (Left, Right : in Natural);
   --  Sorts range Left .. Right of your data by calling 'Is_Smaller' for
   --  comparisons and 'Move' to move elements around. Both 'Is_Smaller'
   --  and 'Move' must be prepared to receive indices -1 and -2, which denote
   --  two single (and different) temporary locations. Both routines are never
   --  called with both indices negative, one index at least is always
   --  in the (Left .. Right).

   ----------------------------------------------------------------------------
   --  Same as above, but using access-to-subroutines.

   type Comparator is access
     function (Left, Right : in Integer) return Boolean;

   type Copier is access
     procedure (To, From : in Integer);

   procedure Sort
     (Left, Right : in Natural;
      Is_Smaller  : in Comparator;
      Copy        : in Copier);
   --  Of course, both 'Is_Smaller' and 'Copy' must not be null or this will
   --  raise Constraint_Error!

   ----------------------------------------------------------------------------

end TW_Sorting;
